import asyncio
import functools
import typing

try:
    # new module from python 3.7+
    # https://docs.python.org/3/whatsnew/3.7.html#contextvars
    import contextvars
except ImportError:
    contextvars = None


_T = typing.TypeVar("_T")


async def run_in_threadpool(
    function: typing.Callable[..., _T], *args: typing.Any, **kwargs: typing.Any
) -> _T:
    """
    run function in another thread inside event loops
    it will not block the whole treads in case when
    function can be some blocking code
    """
    loop = asyncio.get_event_loop()

    if contextvars is not None:
        child = functools.partial(function, *args, **kwargs)
        context = contextvars.copy_context()
        function = context.run
        args = (child,)
    elif kwargs:
        function = functools.partial(function, **kwargs)

    # https://cheat.readthedocs.io/en/latest/python/asyncio.html
    # None uses default ThreadPoolExecutor
    # (might want to give option to support executator `ProcessPoolExecutor`
    #  for CPU intensive tasks)
    return await loop.run_in_executor(None, function, *args)


class _StopGenerator(Exception):
    pass


def _next(iterator: typing.Iterator) -> typing.Any:
    # You can raise anything other than StopIteration inside futures
    # https://www.osgeo.cn/tornado/_modules/asyncio/futures.html
    try:
        return next(iterator)
    except StopIteration:
        raise _StopGenerator


async def generate_in_threadpool(
    iterator: typing.Iterator,
) -> typing.AsyncGenerator:
    while True:
        try:
            yield await run_in_threadpool(_next, iterator)
        except _StopGenerator:
            break


async def run_until_first_complete(
    tasks: typing.Tuple[typing.Callable],
) -> None:
    # create tasks
    # https://docs.python.org/3/library/asyncio-task.html#asyncio-example-wait-coroutine
    try:
        tasks = tuple(asyncio.create_task(task) for task in tasks)
    except AttributeError:
        # create_task is added in python3.7 and in python3.6 it is only available
        # as a method on event loop
        loop = asyncio.get_event_loop()
        tasks = tuple(loop.create_task(task) for task in tasks)

    # run until at least one of the task is complete
    # https://docs.python.org/3/library/asyncio-task.html#waiting-primitives
    done, pending = await asyncio.wait(
        tasks, return_when=asyncio.FIRST_COMPLETED
    )

    # cancel pending tasks
    for task in pending:
        task.cancel()



import http

from yaat.responses import TextResponse


class HTTPException(Exception):
    def __init__(self, status_code: int, details: str = None):
        if details is None:
            details = http.HTTPStatus(status_code).phrase
        self.status_code = status_code
        self.details = details

    @property
    def response(self) -> TextResponse:
        return TextResponse(
            status_code=self.status_code, content=self.details,
        )

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.details!r})"

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}(status_code={self.status_code!r}, detail={self.details!r})"


class WebSocketException(RuntimeError):
    def __init__(self, message: str):
        self.message = message

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name} (Message: {self.message!r})"

    def __str__(self) -> str:
        return self.message



from enum import Enum
from parse import parse
import typing

from yaat.constants import HTTP_METHODS


class RouteTypes(Enum):
    HTTP = 1  # http route
    STATIC = 2  # static handler route
    WEBSOCKET = 3  # websocket route


class Route:
    def __init__(
        self,
        route_type: RouteTypes,
        path: str,
        handler: callable,
        methods: list = None,
    ):
        self.route_type = route_type
        self.path = path
        self.handler = handler
        self.methods = methods if methods else HTTP_METHODS

    @property
    def type(self) -> RouteTypes:
        return self.route_type

    @property
    def methods(self) -> list:
        return self.__methods

    @methods.setter
    def methods(self, methods: list):
        # make sure all HTTP methods are upper
        self.__methods = [method.upper() for method in methods]

    def is_valid_method(self, method: str) -> bool:
        return method.upper() in self.methods


class Router:
    def __init__(self):
        self.routes = []

    @property
    def paths(self) -> typing.List[str]:
        paths = set()
        for route in self.routes:
            paths.add(route.path)
        return list(paths)

    def route(self, path: str, methods: list = None) -> callable:
        def wrapper(handler):
            self.add_route(path=path, handler=handler, methods=methods)
            return handler

        return wrapper

    def add_route(
        self,
        path: str,
        handler: callable,
        methods: list = None,
        is_static: bool = False,
    ):
        assert path not in self.paths, f"Route {path}, already exists"
        route_type = RouteTypes.STATIC if is_static else RouteTypes.HTTP
        self.routes.append(
            Route(
                route_type=route_type,
                path=path,
                handler=handler,
                methods=methods,
            )
        )

    def websocket_route(self, path: str) -> callable:
        def wrapper(handler):
            self.add_websocket_route(path=path)
            return handler

        return wrapper

    def add_websocket_route(self, path: str, handler: callable):
        assert path not in self.paths, f"Route {path}, already exists"
        self.routes.append(
            Route(route_type=RouteTypes.WEBSOCKET, path=path, handler=handler)
        )

    def mount(self, router: callable, prefix: str = None):
        """Mount another router"""
        routes = router.routes

        # register sub routes
        for route in routes:
            if route.type != RouteTypes.STATIC:
                path = (
                    self.__add_prefix(prefix, route.path)
                    if prefix
                    else route.path
                )

                if route.type == RouteTypes.WEBSOCKET:
                    self.add_websocket_route(path=path, handler=route.handler)
                else:
                    self.add_route(
                        path=path, handler=route.handler, methods=route.methods
                    )
            else:
                # StaticFile is not mounted yet to any router
                path = prefix if prefix else "/"
                if route.path is not None:
                    path = self.__add_prefix(path, route.path)

                # update StaticFileHandler path
                router.path = path
                self.add_route(
                    path=path,
                    handler=route.handler,
                    methods=route.methods,
                    is_static=True,
                )

    def get_route(self, request_path: str) -> (Route, typing.Any):
        for route in self.routes:
            # for static routing, use different methods for route comparison
            if route.type == RouteTypes.STATIC and request_path.startswith(
                route.path
            ):
                return route, {}

            parse_result = parse(route.path, request_path)
            if parse_result is not None:
                return route, parse_result.named

        return None, None

    def __add_prefix(self, prefix: str, path: str) -> str:
        if not prefix.startswith("/"):
            prefix = "/" + prefix

        return (
            f"{prefix}{path}" if path.startswith("/") else f"{prefix}/{path}"
        )



# https://asgi.readthedocs.io/en/latest/specs/www.html#wsgi-encoding-differences
ENCODING_METHOD = "latin-1"

HTTP_METHODS = [
    "GET",
    "HEAD",
    "PATCH",
    "POST",
    "PUT",
    "DELETE",
    "CONNECT",
    "OPTIONS",
    "TRACE",
]


class WebSocketMessages:
    # https://asgi.readthedocs.io/en/latest/specs/www.html#websocket
    ACCEPT = "websocket.accept"
    CONNECT = "websocket.connect"
    DISCONNECT = "websocket.disconnect"
    SEND = "websocket.send"
    RECEIVE = "websocket.receive"
    CLOSE = "websocket.close"


class WebSocketCloseEvent:
    # https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent
    NORMAL_CLOSURE = 1000
    GOING_AWAY = 1001
    PROTOCOL_ERROR = 1002
    UNSUPPORTED_DATA = 1003
    NO_STATUS_RECEIVED = 1005
    ABNORMAL_CLOSURE = 1006
    INVALID_FRAME_PAYLOAD_DATA = 1007
    POLICY_VIOLATION = 1008
    MESSAGE_TOO_BIG = 1009
    MISSING_EXTENSION = 1010
    INTERNAL_ERROR = 1011
    SERVICE_RESTART = 1012
    TRY_AGAIN_LATER = 1013
    BAD_GATEWAY = 1014
    TLS_HANDSHAKE = 1015



from multipart.multipart import parse_options_header
import http.cookies
import json
import typing

from yaat.components import Address, Form, Headers, QueryParams, URL
from yaat.constants import ENCODING_METHOD
from yaat.parsers import FormParser, MultiPartParser
from yaat.typing import Scope, Receive, Send, Message


async def empty_receive() -> Message:
    raise RuntimeError("Receive channel has not been made available")


class HTTPConnection:
    def __init__(self, scope: Scope):
        assert scope["type"] in ("http", "websocket")
        self.scope = scope

    def __getitem__(self, key: str) -> str:
        return self.scope[key]

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self.scope)

    def __len__(self) -> int:
        return len(self.scope)

    @property
    def app(self) -> typing.Any:
        return self.scope["app"]

    @property
    def method(self) -> str:
        return self.scope["method"]

    @property
    def path(self) -> str:
        return self.scope["path"]

    @property
    def url(self) -> URL:
        if not hasattr(self, "_url"):
            self._url = URL(scope=self.scope)
        return self._url

    @property
    def headers(self) -> Headers:
        if not hasattr(self, "_headers"):
            self._headers = Headers(self.scope["headers"])
        return self._headers

    @property
    def query_params(self) -> QueryParams:
        if not hasattr(self, "_query_params"):
            self._query_params = QueryParams(self.scope["query_string"])
        return self._query_params

    @property
    def cookies(self) -> typing.Dict[str, str]:
        if not hasattr(self, "_cookies"):
            cookies = {}
            raw_cookies = self.headers.get("cookie")

            if raw_cookies:
                cookie = http.cookies.SimpleCookie()
                cookie.load(raw_cookies)

                for key, data in cookie.items():
                    cookies[key] = data.value

            self._cookies = cookies

        return self._cookies

    @property
    def client(self) -> Address:
        host, port = self.scope.get("client") or (None, None)
        return Address(host, port) if host else None


class Request(HTTPConnection):
    def __init__(
        self, scope: Scope, receive: Receive = empty_receive,
    ):
        super().__init__(scope)
        self.receive = receive
        assert scope["type"] == "http"

    @property
    def receive(self) -> Receive:
        return self.__receive

    @receive.setter
    def receive(self, receive: Receive):
        self.__receive = receive

    async def stream(self) -> typing.AsyncGenerator[bytes, None]:
        if hasattr(self, "_body"):
            yield self._body
            yield b""
            return

        while True:
            message = await self.receive()
            body = message.get("body", b"")
            if body:
                yield body
            if not message.get("more_body", False):
                break

        yield b""

    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)

        return self._body

    async def json(self) -> typing.Any:
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = {}

            if not body == b"":
                self._json = json.loads(body)
        return self._json

    async def form(self) -> dict:
        if not hasattr(self, "_form"):
            content_type, options = parse_options_header(
                self.headers.get("content-type")
            )

            if content_type == b"multipart/form-data":
                multipart_parser = MultiPartParser(self.headers, self.stream())
                self._form = await multipart_parser.parse()

            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.body)
                self._form = await form_parser.parse()

            else:
                self._form = Form()

        return self._form



import httpx
import inspect
import typing

from yaat.exceptions import HTTPException
from yaat.middleware import (
    BaseMiddleware,
    ExceptionMiddleware,
    LifespanMiddleware,
)
from yaat.parsers import UrlParamParser
from yaat.requests import Request
from yaat.responses import Response
from yaat.routing import Router, RouteTypes
from yaat.staticfiles import StaticFilesHandler
from yaat.typing import Scope, Receive, Send
from yaat.websockets import WebSocket


class Yaat:
    def __init__(
        self,
        middlewares: typing.Sequence[BaseMiddleware] = None,
        on_startup: typing.Sequence[callable] = None,
        on_shutdown: typing.Sequence[callable] = None,
    ):
        self.router = Router()
        self.middleware = LifespanMiddleware(
            app=self, on_startup=on_startup, on_shutdown=on_shutdown
        )

        # NOTE: setup middleware(s) registration
        self.__setup_middlewares(middlewares)

    # NOTE: Routing
    def route(self, path: str, methods: list = None) -> callable:
        def wrapper(handler):
            self.add_route(path, handler, methods)
            return handler

        return wrapper

    def add_route(self, path: str, handler: callable, methods: list = None):
        self.router.add_route(path=path, handler=handler, methods=methods)

    def websocket_route(self, path: str) -> callable:
        def wrapper(handler):
            self.add_websocket_route(path, handler)
            return handler

        return wrapper

    def add_websocket_route(self, path: str, handler: callable):
        self.router.add_websocket_route(path=path, handler=handler)

    def mount(self, router: Router, prefix: str = None):
        self.router.mount(router=router, prefix=prefix)

    # NOTE: Handle HTTP Request
    async def handle_request(self, request: Request) -> Response:
        route, kwargs = self.router.get_route(request_path=request.path)

        try:
            # check if route exists and it is not websocket route
            if (
                route
                and route.handler is not None
                and route.type != RouteTypes.WEBSOCKET
            ):
                handler = route.handler

                if inspect.isclass(handler):
                    handler = getattr(handler(), request.method.lower(), None)
                    if handler is None:
                        raise HTTPException(405)
                if not route.is_valid_method(request.method):
                    raise HTTPException(405)

                # convert url param datatypes to annotation types
                param_parser = UrlParamParser(
                    handler, kwargs, inspect.isclass(route.handler)
                )
                kwargs = param_parser.get()

                if isinstance(handler, StaticFilesHandler):
                    # pass route path to static files handler
                    kwargs["router_path"] = route.path
                    response = await handler(request, **kwargs)
                else:
                    response = await handler(request, **kwargs)
            else:
                raise HTTPException(404)
        except Exception as e:
            if isinstance(e, HTTPException) or isinstance(e, HTTPException):
                response = e.response
            else:
                raise e
        return response

    # NOTE: Handle Websocket
    async def handle_websocket(self, websocket: WebSocket):
        route, _ = self.router.get_route(request_path=websocket.path)

        if route and route.handler is not None:
            handler = route.handler
            await handler(websocket)

    # NOTE: Middleware
    def add_middleware(self, middleware_cls: BaseMiddleware, *args, **kwargs):
        self.middleware.add(middleware_cls, *args, **kwargs)

    def __setup_middlewares(
        self, middlewares: typing.Sequence[BaseMiddleware] = None
    ):
        # register exception handling middleware
        self.add_middleware(ExceptionMiddleware)

        # register user defined middlewares
        if not middlewares:
            middlewares = []
        for middleware in middlewares:
            self.add_middleware(middleware)

    # NOTE: Test Client
    def test_client(
        self, base_url: str = "http://testserver"
    ) -> httpx.AsyncClient:
        if not hasattr(self, "_test_client"):
            self._test_client = httpx.AsyncClient(app=self, base_url=base_url)

        return self._test_client

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        await self.middleware(scope, receive, send)



import asyncio
import typing

from yaat.concurrency import run_in_threadpool


class BackgroundTask:
    def __init__(
        self,
        function: typing.Callable,
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        self.function = function
        self.args = args
        self.kwargs = kwargs

    async def __call__(self):
        if asyncio.iscoroutinefunction(self.function):
            await self.function(*self.args, **self.kwargs)
        else:
            await run_in_threadpool(self.function, *self.args, **self.kwargs)


class BackgroundTasks:
    """
    to group individual background tasks and run multiple
    background tasks in single call
    """

    def __init__(self):
        self.tasks = []

    def add(
        self,
        function: typing.Callable,
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        self.tasks.append(BackgroundTask(function, *args, **kwargs))

    async def __call__(self):
        for task in self.tasks:
            await task()



import typing

from yaat.requests import Request
from yaat.responses import Response
from yaat.typing import ASGIApp, Scope, Send, Receive
from yaat.websockets import WebSocket


class BaseMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    def add(self, middleware_cls: typing.Type, *args, **kwargs):
        self.app = middleware_cls(self.app, *args, **kwargs)

    async def process_request(self, request: Request):
        pass

    async def process_response(self, response: Response):
        pass

    async def handle_request(self, request: Request) -> Response:
        await self.process_request(request)
        response = await self.app.handle_request(request)
        await self.process_response(response)
        return response

    async def handle_websocket(self, websocket: WebSocket):
        await self.app.handle_websocket(websocket)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # Handle Websocket
        if scope["type"] == "websocket":
            websocket = WebSocket(scope=scope, receive=receive, send=send)
            await self.handle_websocket(websocket)
        # Handle Request
        else:
            request = Request(scope, receive)
            response = await self.handle_request(request)
            await response(scope, receive, send)



import enum
import json
import typing

from yaat.constants import (
    WebSocketCloseEvent as WsCloseEvent,
    WebSocketMessages as WsMessages,
)
from yaat.exceptions import WebSocketException
from yaat.requests import HTTPConnection
from yaat.typing import Message, Scope, Receive, Send


# CONSTANTS
class WebSocketStates(enum.Enum):
    # To identify whether client/server is connected/disconnected
    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class WebSocketDisconnect(Exception):
    def __init__(self, code: int = WsCloseEvent.NORMAL_CLOSURE):
        self.code = code

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name} (Status Code: {self.code!r})"

    def __str__(self) -> str:
        return f"WebSocket Disconnected with ({self.code!r})"


class WebSocket(HTTPConnection):
    def __init__(self, scope: Scope, receive: Receive, send: Send):
        super().__init__(scope)
        assert scope["type"] == "websocket"

        self.__receive = receive
        self.__send = send
        self.client_state = WebSocketStates.CONNECTING
        self.server_state = WebSocketStates.CONNECTING

    async def receive(self) -> Message:
        """
        Receive ASGI websocket message
        """
        if self.client_state == WebSocketStates.CONNECTING:
            message = await self.__receive()
            assert message["type"] == WsMessages.CONNECT
            self.client_state = WebSocketStates.CONNECTED
            return message

        elif self.client_state == WebSocketStates.CONNECTED:
            message = await self.__receive()
            message_type = message["type"]
            assert message_type in {WsMessages.RECEIVE, WsMessages.DISCONNECT}

            if message_type == WsMessages.DISCONNECT:
                self.cleint_state = WebSocketStates.DISCONNECTED
            return message

        else:
            raise WebSocketException("Cannot 'receive' when disconnected")

    async def send(self, message: Message):
        """
        Send ASGI websocket message
        """
        if self.server_state == WebSocketStates.CONNECTING:
            message_type = message["type"]
            assert message_type in {WsMessages.ACCEPT, WsMessages.CLOSE}

            if message_type == WsMessages.CLOSE:
                self.server_state = WebSocketStates.DISCONNECTED
            else:
                self.server_state = WebSocketStates.CONNECTED
            await self.__send(message)

        elif self.server_state == WebSocketStates.CONNECTED:
            message_type = message["type"]
            assert message_type in {WsMessages.SEND, WsMessages.CLOSE}

            if message_type == WsMessages.CLOSE:
                self.server_state = WebSocketStates.DISCONNECTED
            await self.__send(message)

        else:
            raise WebSocketException("Cannot 'send' when disconnected")

    # Accept/Close connections
    async def accept(self, subprotocol: str = None):
        if self.client_state == WebSocketStates.CONNECTING:
            await self.receive()
        await self.send(
            {"type": WsMessages.ACCEPT, "subprotocol": subprotocol}
        )

    async def close(self, code: int = WsCloseEvent.NORMAL_CLOSURE):
        # Websocket status codes
        # https://developer.mozilla.org/en-US/docs/Web/API/CloseEvent#Status_codes
        await self.send({"type": WsMessages.CLOSE, "code": code})

    # Sending methods
    async def send_text(self, data: str):
        await self.send({"type": WsMessages.SEND, "text": data})

    async def send_bytes(self, data: bytes):
        await self.send({"type": WsMessages.SEND, "bytes": data})

    async def send_json(self, data: typing.Any, mode: str = "text"):
        assert mode in ["text", "bytes"]
        raw = json.dumps(data)

        if mode == "text":
            message = {"type": WsMessages.SEND, "text": raw}
        else:
            message = {"type": WsMessages.SEND, "bytes": raw.encode("utf-8")}

        await self.send(message)

    # Receiving methods
    async def receive_text(self) -> str:
        assert self.server_state == WebSocketStates.CONNECTED
        message = await self.receive()
        self.__raise_if_disconnected(message)
        return message["text"]

    async def receive_bytes(self) -> bytes:
        assert self.server_state == WebSocketStates.CONNECTED
        message = await self.receive()
        self.__raise_if_disconnected(message)
        return message["bytes"]

    async def receive_json(self, mode: str = "text") -> typing.Any:
        assert mode in ["text", "bytes"]
        assert self.server_state == WebSocketStates.CONNECTED

        message = await self.receive()
        self.__raise_if_disconnected(message)
        text = (
            message["text"]
            if mode == "text"
            else message["bytes"].decode("utf-8")
        )
        return json.loads(text)

    # Exception handlers
    def __raise_if_disconnected(self, message: Message):
        if message["type"] == WsMessages.DISCONNECT:
            raise WebSocketDisconnect(message["code"])



from email.utils import formatdate
from mimetypes import guess_type
from urllib.parse import quote, quote_plus
import aiofiles
import hashlib
import http.cookies
import inspect
import json
import os
import typing

from yaat.background import BackgroundTask
from yaat.concurrency import generate_in_threadpool, run_until_first_complete
from yaat.constants import ENCODING_METHOD
from yaat.typing import Scope, Receive, Send


# Workaround for adding samesite support to python versions below 3.8
# https://stackoverflow.com/a/50813092
http.cookies.Morsel._reserved["samesite"] = "SameSite"


class Response:
    media_type = None
    charset = "utf-8"

    def __init__(
        self,
        content: typing.Any = None,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
    ):
        self.status_code = status_code
        if media_type is not None:
            self.media_type = media_type
        self.headers = headers if headers is not None else {}
        self.body = self.render_content(content)

    def render_content(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return content.encode(self.charset)

    def get_raw_headers(
        self, headers: typing.Mapping[str, str] = None
    ) -> list:
        if headers is None:
            headers = self.headers

        if headers is None:
            raw_headers = []  # type: typing.List[typing.Tuple[bytes, bytes]]
            populate_content_length = True
            populate_content_type = True
        else:
            raw_headers = [
                (k.lower().encode(ENCODING_METHOD), v.encode(ENCODING_METHOD))
                for k, v in headers.items()
            ]
            keys = [h[0] for h in raw_headers]
            populate_content_length = b"content-length" not in keys
            populate_content_type = b"content-type" not in keys

        body = getattr(self, "body", b"")
        if body and populate_content_length:
            content_length = str(len(body))
            raw_headers.append(
                (b"content-length", content_length.encode(ENCODING_METHOD))
            )

        content_type = self.media_type
        if content_type is not None and populate_content_type:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            raw_headers.append(
                (b"content-type", content_type.encode(ENCODING_METHOD))
            )

        return raw_headers

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int = None,
        expires: int = None,
        path: str = "/",
        domain: str = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: str = "lax",
    ):
        cookie: http.cookies.BaseCookie = http.cookies.SimpleCookie()
        cookie[key] = value
        if max_age is not None:
            cookie[key]["max-age"] = max_age
        if expires is not None:
            cookie[key]["expires"] = expires
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            # 'none' for cross-site access
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be either 'strict', 'lax' or 'none'"

            cookie[key]["samesite"] = samesite

        self.headers["set-cookie"] = cookie.output(header="").strip()

    def delete_cookie(self, key: str, path: str = "/", domain: str = None):
        self.set_cookie(
            key=key, path=path, domain=domain, expires=0, max_age=0
        )

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.get_raw_headers(),
            }
        )
        await send({"type": "http.response.body", "body": self.body})


class HTMLResponse(Response):
    media_type = "text/html"


class TextResponse(Response):
    media_type = "text/plain"


class JSONResponse(Response):
    media_type = "application/json"

    def render_content(self, content: typing.Any) -> bytes:
        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
        ).encode("utf-8")


class RedirectResponse(Response):
    def __init__(self, url: str, status_code: int = 307, headers: dict = None):
        if not headers:
            headers = {}

        headers["location"] = quote_plus(str(url), safe=":/%#?&=@[]!$&'()*+,;")
        super().__init__(content=b"", status_code=status_code, headers=headers)


class FileResponse(Response):
    chunk_size = 4096

    def __init__(
        self,
        path: str,
        filename: str = None,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        stat_result: os.stat_result = None,
        method: str = None,
    ):
        self.path = path
        self.filename = filename
        self.status_code = status_code
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if media_type is None:
            media_type = guess_type(filename or str(path))[0] or "text/plain"
        self.headers = headers if headers is not None else {}
        self.media_type = media_type
        self.stat_result = stat_result

        if self.filename is not None:
            content_disposition_filename = quote(self.filename)

            if content_disposition_filename != self.filename:
                content_disposition = "attachment; filename*=utf-8''{}".format(
                    content_disposition_filename
                )
            else:
                content_disposition = 'attachment; filename="{}"'.format(
                    self.filename
                )
            self.headers["content-disposition"] = content_disposition

        if stat_result is not None:
            self.set_stat_headers(stat_result)

    def set_stat_headers(self, stat_result: os.stat_result):
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = hashlib.md5(etag_base.encode()).hexdigest()

        self.headers["content-length"] = content_length
        self.headers["last-modified"] = last_modified
        self.headers["etag"] = etag

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        # Send 404 if file does not exists
        if not os.path.exists(self.path):
            # clear custom headers
            await send(
                {
                    "type": "http.response.start",
                    "status": 404,
                    "headers": self.get_raw_headers(
                        headers={"content-type": "text/plain"}
                    ),
                }
            )
            await send(
                {"type": "http.response.body", "body": b"file not found"}
            )
            return

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.get_raw_headers(),
            }
        )
        if self.send_header_only:
            await send({"type": "http.response.body"})
        else:
            async with aiofiles.open(self.path, mode="rb") as file:
                more_body = True
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    more_body = len(chunk) == self.chunk_size
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )


class NotModifiedResponse(Response):
    """ Use when the file requested from user is not modifed
         so instead of sending FileResponse, send NotModifiedResponse with
         empty body """

    NOT_MODIFIED_HEADERS = (
        "cache-control",
        "content-location",
        "date",
        "etag",
        "expires",
        "vary",
    )

    def __init__(self, headers: dict):
        headers = {
            name: value
            for name, value in headers.items()
            if name in self.NOT_MODIFIED_HEADERS
        }
        super().__init__(status_code=304, headers=headers)


class StreamResponse(Response):
    def __init__(
        self,
        content: typing.Any,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
    ):
        super().__init__(
            status_code=status_code, headers=headers, media_type=media_type
        )
        # check if async generator, if not create generator
        if inspect.isasyncgen(content):
            self.body_gen = content
        else:
            self.body_gen = generate_in_threadpool(content)

        # NOTE: workaround for pytest client as it test client continue
        #        listening for request even after streaming is completed
        self.streaming = False

    async def when_disconnect_or_finish(self, receive: Receive):
        while True:
            message = await receive()
            if message["type"] == "http.disconnect" or not self.streaming:
                self.streaming = False
                break

    async def stream(self, send: Send):
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.get_raw_headers(),
            }
        )

        async for chunk in self.body_gen:
            await send(
                {
                    "type": "http.response.body",
                    "body": chunk
                    if isinstance(chunk, bytes)
                    else chunk.encode(self.charset),
                    "more_body": True,
                }
            )

        await send(
            {"type": "http.response.body", "body": b"", "more_body": False}
        )
        self.streaming = False

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        self.streaming = True
        tasks = (self.stream(send), self.when_disconnect_or_finish(receive))
        await run_until_first_complete(tasks)


class RunAfterResponse:
    """ Use to return response and run background task(s) after """

    def __init__(self, response: Response, background: BackgroundTask):
        self.response = response
        self.background = background

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        await self.response(scope, receive, send)
        await self.background()



import re
import typing

from yaat.components import Headers
from yaat.constants import HTTP_METHODS
from yaat.middleware import BaseMiddleware
from yaat.requests import Request
from yaat.responses import Response, TextResponse
from yaat.typing import ASGIApp


# https://developer.mozilla.org/en-US/docs/Glossary/CORS-safelisted_request_header
SAFELISTED_HEADERS = {
    "Accept",
    "Accept-Language",
    "Content-Language",
    "Content-Type",
}


class CORSMiddleware(BaseMiddleware):
    def __init__(
        self,
        app: ASGIApp,
        allow_origins: typing.Sequence[str] = (),
        allow_origin_regex: str = None,
        allow_methods: typing.Sequence[str] = ("GET",),
        allow_headers: typing.Sequence[str] = (),
        allow_credentials: bool = False,
        expose_headers: typing.Sequence[str] = (),
        max_age: int = 600,  # 10 mins
    ):
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS

        super().__init__(app)

        if "*" in allow_methods:
            allow_methods = HTTP_METHODS

        self.allow_origins = allow_origins
        self.allow_methods = allow_methods
        self.allow_headers = [header.lower() for header in allow_headers]
        self.allow_credentials = allow_credentials
        self.expose_headers = expose_headers
        self.max_age = max_age
        self.allow_all_origins = "*" in allow_origins
        self.allow_all_headers = "*" in allow_headers
        self.allow_origin_regex = (
            re.compile(allow_origin_regex) if allow_origin_regex else None
        )
        self.preflight_allow_headers = sorted(
            SAFELISTED_HEADERS | set(allow_headers)
        )

    def is_allowed_origin(self, origin: str) -> bool:
        if self.allow_all_origins:
            return True

        if self.allow_origin_regex and self.allow_origin_regex.fullmatch(
            origin
        ):
            return True

        return origin in self.allow_origins

    def preflight_response(self, request_headers: Headers) -> Response:
        # https://developer.mozilla.org/en-US/docs/Glossary/Preflight_request

        requested_origin = request_headers["origin"]
        requested_method = request_headers["access-control-request-method"]
        requested_headers = request_headers.get(
            "access-control-request-headers"
        )

        # preflight headers
        headers = {
            "Access-Control-Allow-Methods": ", ".join(self.allow_methods),
            "Access-Control-Allow-Headers": ", ".join(
                self.preflight_allow_headers
            ),
            "Access-Control-Max-Age": str(self.max_age),
        }
        if self.allow_credentials:
            headers["Access-Control-Allow-Credentials"] = "true"
        failures = []

        # check origin
        if (
            self.is_allowed_origin(requested_origin)
            and not self.allow_all_origins
        ):
            headers["Access-Control-Allow-Origin"] = requested_origin
            headers["Vary"] = "Origin"
        elif self.allow_all_origins:
            headers["Access-Control-Allow-Origin"] = "*"
        else:
            failures.append("Origin")

        # check HTTP method
        if requested_method not in self.allow_methods:
            failures.append("Method")

        # check allow headers
        if requested_headers is not None and self.allow_all_headers:
            headers["Access-Control-Allow-Headers"] = requested_headers
        elif requested_headers is not None:
            for header in requested_headers.split(","):
                if header.lower().strip() not in self.allow_headers:
                    failures.append("Headers")
                    break

        if failures:
            message = f"Disallowed cors {', '.join(failures)}"
            return TextResponse(message, status_code=400, headers=headers)

        # do not return 204, as some legacy browser does not work with 204
        return Response(status_code=200, headers=headers)

    def simple_response(
        self, request_headers: Headers, response: Response
    ) -> Response:
        origin = request_headers["origin"]
        has_cookie = "cookie" in request_headers

        # check allow credentials
        if self.allow_credentials:
            response.headers["Access-Control-Allow-Credentials"] = "true"

        # check expose header
        if self.expose_headers:
            response.headers["Access-Control-Expose-Headers"] = ", ".join(
                self.expose_headers
            )

        # check origin
        # if cookie presents, server must specify the origin
        # instead of wildcard
        if self.allow_all_origins and has_cookie:
            response.headers["Access-Control-Allow-Origin"] = origin

        elif self.is_allowed_origin(origin) and not self.allow_all_origins:
            response.headers["Access-Control-Allow-Origin"] = origin

            vary = response.headers.get("Vary")
            if vary:
                vary = vary.split(",")
                vary.append("Origin")
            else:
                vary = ["Origin"]
            vary = ", ".join(vary)
            response.headers["Vary"] = vary

        elif self.allow_all_origins:
            response.headers["Access-Control-Allow-Origin"] = "*"

        return response

    async def handle_request(self, request: Request) -> Response:
        origin = request.headers.get("origin")

        # not cors
        if not origin:
            return await self.app.handle_request(request)

        # preflight request
        if (
            request.method == "OPTIONS"
            and "access-control-request-method" in request.headers
        ):
            return self.preflight_response(request.headers)

        response = await self.app.handle_request(request)
        return self.simple_response(request.headers, response)



import traceback

from yaat.exceptions import HTTPException, WebSocketException
from yaat.middleware.base import BaseMiddleware
from yaat.requests import Request
from yaat.responses import Response
from yaat.websockets import WebSocket, WebSocketDisconnect


class ExceptionMiddleware(BaseMiddleware):
    async def handle_request(self, request: Request) -> Response:
        try:
            response = await self.app.handle_request(request)
        except Exception:
            # print traceback and send 500 Internal Server Error
            traceback.print_exc()
            response = HTTPException(500).response

        return response

    async def handle_websocket(self, websocket: WebSocket):
        try:
            await self.app.handle_websocket(websocket)

        except WebSocketDisconnect as ws_err:
            raise ws_err

        except WebSocketException as ws_err:
            traceback.print_exc()
            raise ws_err

        except Exception:
            traceback.print_exc()



from setuptools import setup

VERSION = "0.1.6"


def get_long_description():
    with open("README.md", encoding="utf8") as f:
        return f.read()


setup(
    name="yaat",
    packages=["yaat", "yaat.middleware"],
    version=VERSION,
    license="LGPL",
    description="Yet another ASGI toolkit",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Khant",
    author_email="contact@khant.dev",
    url="https://github.com/yaat-project/yaat",
    download_url=f"https://github.com/yaat-project/yaat/releases/download/v{VERSION}/yaat-{VERSION}.tar.gz",
    keywords=["asynchronous", "web framework"],
    python_requires=">=3.6",
    install_requires=[
        "aiofiles",
        "httpx",
        "Jinja2",
        "parse",
        "python-multipart",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import asyncio
import traceback
import typing

from yaat.middleware.base import BaseMiddleware
from yaat.typing import ASGIApp, Scope, Send, Receive


class LifespanMiddleware(BaseMiddleware):
    """
    Middleware to handle ASGI startup and shutdown lifespan events.
    """

    def __init__(
        self,
        app: ASGIApp,
        on_startup: typing.Sequence[callable] = None,
        on_shutdown: typing.Sequence[callable] = None,
    ):
        super().__init__(app)
        self.on_startup = on_startup if on_startup else []
        self.on_shutdown = on_shutdown if on_shutdown else []

    async def startup(self):
        for method in self.on_startup:
            if asyncio.iscoroutinefunction(method):
                await method()
            else:
                method()

    async def shutdown(self):
        for method in self.on_shutdown:
            if asyncio.iscoroutinefunction(method):
                await method()
            else:
                method()

    async def lifespan(self, receive: Receive, send: Send):
        message = await receive()
        assert message["type"] == "lifespan.startup"

        try:
            await self.startup()
        except Exception:
            message = traceback.print_exc()
            await send({"type": "lifespan.startup.failed", "message": message})
            raise

        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"

        try:
            await self.shutdown()
        except Exception:
            traceback.print_exc()

        await self.shutdown()
        await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        assert scope["type"] in ("http", "websocket", "lifespan")

        if scope["type"] == "lifespan":
            await self.lifespan(receive, send)
        else:
            await super().__call__(scope, receive, send)



from yaat.middleware.base import BaseMiddleware
from yaat.middleware.cors import CORSMiddleware
from yaat.middleware.exception import ExceptionMiddleware
from yaat.middleware.lifespan import LifespanMiddleware
